/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2012 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.graphicsGL.ui.objects;

import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.drawLine;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.drawMaterialGroups;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.flushName;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.moveTo;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.prepareName;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.restoreMatrix;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.saveMatrix;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.setColor;
import static org.fortiss.tooling.graphicsGL.util.Graphics3DModelElementFactory.BLUE;
import static org.fortiss.tooling.graphicsGL.util.Graphics3DModelElementFactory.WHITE;
import static org.lwjgl.util.glu.GLU.GLU_FILL;

import java.io.File;
import java.io.IOException;
import java.nio.FloatBuffer;
import java.util.List;

import org.fortiss.tooling.graphicsGL.model.Color3D;
import org.fortiss.tooling.graphicsGL.model.graph.Axis3D;
import org.fortiss.tooling.graphicsGL.model.graph.CoordinateSystem3D;
import org.fortiss.tooling.graphicsGL.model.graph.Graph3D;
import org.fortiss.tooling.graphicsGL.model.graph.GraphLine3D;
import org.fortiss.tooling.graphicsGL.model.graph.GraphPoint3D;
import org.fortiss.tooling.graphicsGL.model.scene.Object3D;
import org.fortiss.tooling.graphicsGL.ui.camera.Camera;
import org.fortiss.tooling.graphicsGL.ui.picker.ModelObjectPicker;
import org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives;
import org.fortiss.tooling.graphicsGL.util.OBJLoaderUtil;
import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.GL11;
import org.lwjgl.util.glu.Sphere;
import org.lwjgl.util.vector.Vector3f;
import org.lwjgl.util.vector.Vector4f;

/**
 * View object for {@link Graph3D}.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 35C66DEC44CFCE044C427F14D4854E72
 */
public class Graph3DViewObject extends ViewObjectBase {
	/** The graph model. */
	private final Graph3D graph;

	/** The highlighted model object. */
	private GraphPoint3D highlighted;

	/** for testing text rendering */
	private List<Object3D> testText;

	/** Constructor. */
	public Graph3DViewObject(Graph3D graph) {
		this.graph = graph;

		File objFile = new File("C:\\Users\\eder\\Desktop\\TestText.obj");
		try {
			OBJLoaderUtil obj = new OBJLoaderUtil(objFile);
			testText = obj.getLoadedObjects();

		} catch(IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
	}

	/** Sets the currently highlighted point. */
	public void setHighlightedPoint(GraphPoint3D point) {
		highlighted = point;
	}

	/** {@inheritDoc} */
	@Override
	public void renderObject(Camera c) {
		renderCoordSys(graph.getCoordinateSystem());
		setColor(WHITE);
		for(GraphPoint3D p : graph.getPointsList()) {
			if(p == highlighted) {
				setColor(BLUE);
				renderPoint(p);
				setColor(WHITE);
			} else {
				setColor(WHITE);
				Color3D color = p.getColor();
				if(color != null) {
					setColor(color);
				}
				renderPoint(p);
				setColor(WHITE);
			}
		}
		for(GraphLine3D l : graph.getLinesList()) {
			drawLine(l.getStart().getPosition(), l.getEnd().getPosition());
		}

		renderText(c);
	}

	/**
	 * render Text
	 * 
	 * @param c
	 */
	private void renderText(Camera c) {

		if(testText == null)
			return;
		saveMatrix();

		Vector3f look = new Vector3f();
		Vector3f pos = new Vector3f(0, 10, 0);
		Vector3f right = new Vector3f();
		Vector3f up;

		// look is now the vector from the camera position to the text position
		// billboarding goes along z axis
		Vector3f.sub(c.getPosition(), pos, look);
		look.z = 0;
		look = normalize(look);

		up = new Vector3f(0, 0, 1);
		Vector3f.cross(up, look, right);

		// set up transformation matrix for text / billboard
		float[] bbmat = new float[16];
		createBillboardMatrix(bbmat, right, up, look, pos);

		FloatBuffer fb = BufferUtils.createFloatBuffer(16);
		fb.put(bbmat);
		fb.flip();

		// apply the billboard
		GL11.glMultMatrix(fb);

		// rotate text so it is readable from left to right
		GLPrimitives.rotate(new Vector4f(0, 0, 1, -90));

		// draw the Text
		for(Object3D o : testText) {
			drawMaterialGroups(o.getMesh().getMaterialGroupsList());
		}
		restoreMatrix();

	}

	/** creates a transformation matrix for billboards. Result is stored in bbmat */
	private void createBillboardMatrix(float[] bbmat, Vector3f right, Vector3f up, Vector3f look,
			Vector3f pos) {

		bbmat[0] = right.x;
		bbmat[1] = right.y;
		bbmat[2] = right.z;
		bbmat[3] = 0;
		bbmat[4] = up.x;
		bbmat[5] = up.y;
		bbmat[6] = up.z;
		bbmat[7] = 0;
		bbmat[8] = look.x;
		bbmat[9] = look.y;
		bbmat[10] = look.z;
		bbmat[11] = 0;
		// Add the translation in as well.
		bbmat[12] = pos.x;
		bbmat[13] = pos.y;
		bbmat[14] = pos.z;
		bbmat[15] = 1;

	}

	/** returns the normalized vector */
	private Vector3f normalize(Vector3f v) {
		float length = (float)Math.sqrt(v.x * v.x + v.y * v.y + v.z * v.z);
		v.x /= length;
		v.y /= length;
		v.z /= length;
		return v;

	}

	/** {@inheritDoc} */
	@Override
	public void renderObjectForPicking(ModelObjectPicker picker, Camera c) {
		// only points are pickable
		for(GraphPoint3D p : graph.getPointsList()) {
			prepareName(p, picker);
			renderPoint(p);
			flushName();
		}
	}

	/** The GLU sphere used for {@link GraphPoint3D} rendering. */
	private static Sphere SPHERE = new Sphere();

	static {
		SPHERE.setDrawStyle(GLU_FILL);
	}

	/** Renders the given point. */
	private void renderPoint(GraphPoint3D p) {
		saveMatrix();
		moveTo(p.getPosition());
		SPHERE.draw(0.2f, 9, 4);
		restoreMatrix();
	}

	/** Renders the coordinate system if present. */
	private void renderCoordSys(CoordinateSystem3D coordinateSystem) {
		if(coordinateSystem != null) {
			if(coordinateSystem.getXAxis() != null) {
				Axis3D a = coordinateSystem.getXAxis();
				renderAxis(a.getColor(), a.getLength(), 0, 0);
			}
			if(coordinateSystem.getYAxis() != null) {
				Axis3D a = coordinateSystem.getYAxis();
				renderAxis(a.getColor(), 0, a.getLength(), 0);
			}
			if(coordinateSystem.getZAxis() != null) {
				Axis3D a = coordinateSystem.getZAxis();
				renderAxis(a.getColor(), 0, 0, a.getLength());
			}
		}
	}

	/** Renders the axis. */
	private void renderAxis(Color3D color, float x, float y, float z) {
		setColor(color);
		drawLine(0, 0, 0, x, y, z);
	}
}
